"""Provide an asynchronous equivalent to the python console."""

import sys
import random
import asyncio
import argparse

from . import code


class AsynchronousCli(code.AsynchronousConsole):

    def __init__(self, commands, streams=None, *, prog=None, loop=None):
        super().__init__(streams=streams, loop=loop)
        self.prog = prog
        self.commands = dict(commands)
        self.commands['help'] = (
            self.help_command,
            argparse.ArgumentParser(
                description='Display the help message.'))
        self.commands['list'] = (
            self.list_command,
            argparse.ArgumentParser(
                description='Display the command list.'))
        self.commands['exit'] = (
            self.exit_command,
            argparse.ArgumentParser(
                description='Exit the interface.'))
        for key, (corofunc, parser) in self.commands.items():
            parser.prog = key
            parser.print_help = lambda file=sys.stderr, *, self=parser: \
                type(parser).print_help(self, file)

    def get_default_banner(self):
        prog = self.prog or sys.argv[0].split('/')[-1]
        msg = "Welcome to the CLI interface of {0}!\n".format(prog)
        msg += "Try:\n"
        msg += " * 'help' to display the help message\n"
        msg += " * 'list' to display the command list."
        return msg

    @asyncio.coroutine
    def help_command(self, reader, writer):
        return """\
Type 'help' to display this message.
Type 'list' to display the command list.
Type '<command> -h' to display the help message of <command>."""

    @asyncio.coroutine
    def list_command(self, reader, writer):
        msg = 'List of commands:'
        for key, (corofunc, parser) in sorted(self.commands.items()):
            usage = parser.format_usage().replace('usage: ', '')[:-1]
            msg += '\n * ' + usage
        return msg

    @asyncio.coroutine
    def exit_command(self, reader, writer):
        raise SystemExit

    @asyncio.coroutine
    def runsource(self, source, filename=None):
        if source.strip().endswith('\\'):
            return True
        source = source.replace('\\\n', '')
        try:
            name, *args = source.split()
        except ValueError:
            return False
        if name not in self.commands:
            self.write("Command '{0}' does not exist.\n".format(name))
            yield from self.flush()
            return False
        corofunc, parser = self.commands[name]
        try:
            namespace = parser.parse_args(args)
        except SystemExit:
            return False
        coro = corofunc(self.reader, self.writer, **vars(namespace))
        try:
            result = yield from coro
        except SystemExit:
            raise
        except:
            self.showtraceback()
        else:
            if result is not None:
                self.write(str(result) + '\n')
        yield from self.flush()
        return False



"""Provide an asynchronous equivalent *to exec*."""

import ast
import codeop
import asyncio

from . import compat

CORO_NAME = '__corofn'
CORO_DEF = 'def {}(): '.format(CORO_NAME)
if compat.PY35:
    CORO_DEF = 'async ' + CORO_DEF
CORO_CODE = CORO_DEF + 'return (None, locals())\n'


def make_arg(key, annotation=None):
    """Make an ast function argument."""
    arg = ast.arg(key, annotation)
    arg.lineno, arg.col_offset = 0, 0
    return arg


def full_update(dct, values):
    """Fully update a dictionary."""
    dct.clear()
    dct.update(values)


def exec_result(obj, local, stream):
    """Reproduce default exec behavior (print and builtins._)"""
    local['_'] = obj
    if obj is not None:
        print(obj, file=stream)


def make_tree(statement, filename="<aexec>", symbol="single", local={}):
    """Helper for *aexec*."""
    # Create tree
    tree = ast.parse(CORO_CODE, filename, symbol)
    # Check expression statement
    if isinstance(statement, ast.Expr):
        tree.body[0].body[0].value.elts[0] = statement.value
    else:
        tree.body[0].body.insert(0, statement)
    # Check and return coroutine
    exec(compile(tree, filename, symbol))
    return tree


def make_coroutine_from_tree(tree, filename="<aexec>", symbol="single",
                             local={}):
    """Make a coroutine from a tree structure."""
    dct = {}
    tree.body[0].args.args = list(map(make_arg, local))
    exec(compile(tree, filename, symbol), dct)
    return asyncio.coroutine(dct[CORO_NAME])(**local)


def compile_for_aexec(source, filename="<aexec>", mode="single",
                      dont_imply_dedent=False, local={}):
    """Return a list of (coroutine object, abstract base tree)."""
    flags = ast.PyCF_ONLY_AST
    if dont_imply_dedent:
        flags |= codeop.PyCF_DONT_IMPLY_DEDENT
    if compat.PY35:
        # Avoid a syntax error by wrapping code with `async def`
        indented = '\n'.join(line and ' ' * 4 + line
                             for line in source.split('\n'))
        coroutine = CORO_DEF + '\n' + indented + '\n'
        interactive = compile(coroutine, filename, mode, flags).body[0]
        # Check EOF errors
        try:
            compile(source, filename, mode, flags)
        except SyntaxError as exc:
            if exc.msg == 'unexpected EOF while parsing':
                raise
    else:
        interactive = compile(source, filename, mode, flags)
    return [make_tree(statement, filename, mode) for
            statement in interactive.body]


@asyncio.coroutine
def aexec(source, local=None, stream=None):
    """Asynchronous equivalent to *exec*.

    Support the *yield from* syntax.
    """
    if local is None:
        local = {}
    if isinstance(source, str):
        source = compile_for_aexec(source)
    for tree in source:
        coro = make_coroutine_from_tree(tree, local=local)
        result, new_local = yield from coro
        exec_result(result, new_local, stream)
        full_update(local, new_local)



"""Provide the apython script."""

import os
import sys
import runpy
import ctypes
import signal
import argparse
import subprocess

from . import events
from . import server

DESCRIPTION = """\
Run the given python file or module with a modified asyncio policy replacing
the default event loop with an interactive loop.
If no argument is given, it simply runs an asynchronous python console."""


def parse_args(args=None):
    parser = argparse.ArgumentParser(prog='apython', description=DESCRIPTION)
    parser.add_argument(
        '--serve', '-s', metavar='[HOST:] PORT',
        help='serve a console on the given interface instead')
    parser.add_argument(
        '--module', '-m', dest='module', action='store_true',
        help='run a python module')
    parser.add_argument(
        '--no-readline', dest='readline', action='store_false',
        help='Force readline disabling')
    parser.add_argument(
        'filename', metavar='FILE', nargs='?',
        help='python file or module to run')
    parser.add_argument(
        'args', metavar='ARGS', nargs=argparse.REMAINDER,
        help='extra arguments')
    namespace = parser.parse_args(args)
    if namespace.module and not namespace.filename:
        parser.error('A python module is required.')
    if namespace.serve is not None:
        namespace.serve = server.parse_server(namespace.serve, parser)
    return namespace


def run_apython(args=None):
    namespace = parse_args(args)

    try:
        import readline
        import rlcompleter
    except ImportError:
        readline, rlcompleter = None, None

    if readline and namespace.readline and not namespace.serve:
        if rlcompleter:
            readline.parse_and_bind("tab: complete")
        code = run_apython_in_subprocess(args)
        sys.exit(code)

    try:
        sys._argv = sys.argv
        sys._path = sys.path
        if namespace.module:
            sys.argv = [None] + namespace.args
            sys.path.insert(0, '')
            events.set_interactive_policy(serve=namespace.serve)
            runpy.run_module(namespace.filename,
                             run_name='__main__',
                             alter_sys=True)
        elif namespace.filename:
            sys.argv = [None] + namespace.args
            path = os.path.dirname(os.path.abspath(namespace.filename))
            sys.path.insert(0, path)
            events.set_interactive_policy(serve=namespace.serve)
            runpy.run_path(namespace.filename,
                           run_name='__main__')
        else:
            events.run_console(serve=namespace.serve)
    finally:
        sys.argv = sys._argv
        sys.path = sys._path


def run_apython_in_subprocess(args=None):
    # Get arguments
    if args is None:
        args = sys.argv[1:]

    # Create subprocess
    proc_args = [sys.executable, '-m', 'aioconsole', '--no-readline']
    process = subprocess.Popen(
        proc_args + args,
        bufsize=0,
        universal_newlines=True,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE)

    # Loop over prompts
    while process.poll() is None:
        try:
            prompt = wait_for_prompt(process.stderr, sys.stderr)
            raw = input_with_stderr_prompt(prompt) + '\n'
        except KeyboardInterrupt:
            process.send_signal(signal.SIGINT)
        except EOFError:
            process.stdin.close()
        else:
            process.stdin.write(raw)

    # Clean up
    sys.stderr.write(process.stderr.read())
    return process.returncode


def wait_for_prompt(src, dest, targets='.>', current='\n'):

    # Read exactly one byte
    def read_one():
        value = src.read(1)
        if value:
            return value
        raise EOFError

    while True:
        # Prompt detection
        if current.endswith('\n'):
            current = reference = read_one()
            if reference in targets:
                while current.endswith(reference):
                    current += read_one()
            if len(current) > 1 and current.endswith(' '):
                return current
        # Regular read
        else:
            current = read_one()
        # Write
        dest.write(current)


def input_with_stderr_prompt(prompt=''):
    api = ctypes.pythonapi
    # Get standard streams
    try:
        fin = ctypes.c_void_p.in_dll(api, 'stdin')
        ferr = ctypes.c_void_p.in_dll(api, 'stderr')
    # Cygwin fallback
    except ValueError:
        raise
        return input(prompt)
    # Call readline
    call_readline = api.PyOS_Readline
    call_readline.restype = ctypes.c_char_p
    result = call_readline(fin, ferr, prompt.encode())
    # Decode result
    if len(result) == 0:
        raise EOFError
    return result.decode().rstrip('\n')


if __name__ == '__main__':
    run_apython()



"""Provide an asychronous equivalent to *input*."""

import io
import sys
import asyncio

from . import compat


class StandardStreamReaderProtocol(asyncio.StreamReaderProtocol):
    def connection_made(self, transport):
        if self._stream_reader._transport is not None:
            return
        super().connection_made(transport)


class StandardStreamReader(asyncio.StreamReader):
    if compat.PY34:
        def __del__(self):
            if self._transport and self._transport.get_extra_info('pipe').fileno() < 3:
                self._transport._pipe = None


class StandardStreamWriter(asyncio.StreamWriter):
    if compat.PY34:
        def __del__(self):
            if self._transport and self._transport.get_extra_info('pipe').fileno() < 3:
                self._transport._pipe = None

    def write(self, data):
        if isinstance(data, str):
            data = data.encode()
        super().write(data)


class NonFileStreamReader:

    def __init__(self, stream, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.stream = stream
        self.eof = False

    def at_eof(self):
        return self.eof

    @asyncio.coroutine
    def readline(self):
        data = yield from self.loop.run_in_executor(None, self.stream.readline)
        if isinstance(data, str):
            data = data.encode()
        self.eof = not data
        return data

    @asyncio.coroutine
    def read(self, n=-1):
        data = yield from self.loop.run_in_executor(None, self.stream.read, n)
        if isinstance(data, str):
            data = data.encode()
        self.eof = not data
        return data

    if compat.PY35:
        @asyncio.coroutine
        def __aiter__(self):
            return self

        @asyncio.coroutine
        def __anext__(self):
            val = yield from self.readline()
            if val == b'':
                raise StopAsyncIteration
            return val


class NonFileStreamWriter:

    def __init__(self, stream, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.stream = stream

    def write(self, data):
        if isinstance(data, bytes):
            data = data.decode()
        self.stream.write(data)

    @asyncio.coroutine
    def drain(self):
        try:
            flush = self.stream.flush
        except AttributeError:
            pass
        else:
            yield from self.loop.run_in_executor(None, flush)


@asyncio.coroutine
def open_pipe_connection(pipe_in, pipe_out, *, loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    reader = StandardStreamReader(loop=loop)
    protocol = StandardStreamReaderProtocol(reader, loop=loop)
    yield from loop.connect_read_pipe(lambda: protocol, pipe_in)
    write_connect = loop.connect_write_pipe(lambda: protocol, pipe_out)
    transport, _ = yield from write_connect
    loop.remove_reader(transport.get_extra_info('pipe').fileno())
    writer = StandardStreamWriter(transport, protocol, reader, loop)
    return reader, writer


@asyncio.coroutine
def create_standard_streams(stdin, stdout, loop):
    try:
        if sys.platform == 'win32':
            raise OSError
        sys.stdin.fileno(), sys.stdout.fileno()
    except OSError:
        reader = NonFileStreamReader(stdin, loop=loop)
        writer = NonFileStreamWriter(stdout, loop=loop)
    else:
        future = open_pipe_connection(stdin, stdout, loop=loop)
        reader, writer = yield from future
    return reader, writer


@asyncio.coroutine
def get_standard_streams(*, cache={}, use_stderr=False, loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    stdin, stdout = sys.stdin, sys.stderr if use_stderr else sys.stdout
    key = loop, stdin, stdout
    if cache.get(key) is None:
        connection = create_standard_streams(stdin, stdout, loop=loop)
        cache[key] = yield from connection
    return cache[key]


@asyncio.coroutine
def ainput(prompt=None, *, loop=None):
    """Asynchronous equivalent to *input*."""
    if loop is None:
        loop = asyncio.get_event_loop()
    if prompt is None:
        prompt = ''
    # Get streams
    reader, writer = yield from get_standard_streams(loop=loop)
    # Write prompt
    writer.write(prompt.encode())
    yield from writer.drain()
    # Get data
    data = (yield from reader.readline()).decode()
    # Return or raise EOF
    if not data.endswith('\n'):
        raise EOFError
    return data.rstrip('\n')



"""Provide an asynchronous equivalent to the python console."""

import sys
import code
import pydoc
import codeop
import signal
import asyncio
import functools
import traceback

from . import stream
from . import compat
from . import execute

EXTRA_MESSAGE = """\
---
This console is running in an asyncio event loop.
It allows you to wait for coroutines using the '{0}' syntax.
Try: {0} asyncio.sleep(1, result=3)
---""".format('await' if compat.PY35 else 'yield from')


class AsynchronousCompiler(codeop.CommandCompiler):

    def __init__(self):
        self.compiler = functools.partial(
            execute.compile_for_aexec,
            dont_imply_dedent=True)


class AsynchronousConsole(code.InteractiveConsole):

    def __init__(self, streams=None, locals=None, filename="<console>",
                 *, loop=None):
        super().__init__(locals, filename)
        self.compile = AsynchronousCompiler()
        if loop is None:
            loop = asyncio.get_event_loop()
        if streams is None:
            self.streams = stream.get_standard_streams(use_stderr=True,
                                                       loop=loop)
        elif isinstance(streams, tuple):
            self.streams = asyncio.coroutine(lambda: streams)()
        else:
            self.streams = streams
        self.reader = None
        self.writer = None
        self.loop = loop
        self.locals.setdefault('asyncio', asyncio)
        self.locals.setdefault('loop', self.loop)
        self.locals.setdefault('print', self.print)
        self.locals.setdefault('help', self.help)

    @functools.wraps(print)
    def print(self, *args, **kwargs):
        kwargs.setdefault('file', self)
        print(*args, **kwargs)

    @functools.wraps(help)
    def help(self, obj):
        self.print(pydoc.render_doc(obj))

    def get_default_banner(self):
        cprt = ('Type "help", "copyright", "credits" '
                'or "license" for more information.')
        msg = "Python %s on %s\n%s\n%s"
        return msg % (sys.version, sys.platform, cprt, EXTRA_MESSAGE)

    @asyncio.coroutine
    def runsource(self, source, filename="<ainput>", symbol="single"):
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(filename)
            return False

        if code is None:
            return True

        yield from self.runcode(code)
        return False

    @asyncio.coroutine
    def runcode(self, code):
        try:
            yield from execute.aexec(code, self.locals, self)
        except SystemExit:
            raise
        except:
            self.showtraceback()
        yield from self.flush()

    def resetbuffer(self):
        self.buffer = []

    def handle_sigint(self, task):
        task.cancel()
        if task._fut_waiter._loop is not self.loop:
            task._wakeup(task._fut_waiter)

    def add_sigint_handler(self):
        task = asyncio.Task.current_task(loop=self.loop)
        try:
            self.loop.add_signal_handler(
                signal.SIGINT, self.handle_sigint, task)
        except NotImplementedError:
            def callback(*args):
                self.loop.call_soon_threadsafe(self.handle_sigint, task)
            signal.signal(signal.SIGINT, callback)

    def remove_sigint_handler(self):
        try:
            self.loop.remove_signal_handler(signal.SIGINT)
        except NotImplementedError:
            signal.signal(signal.SIGINT, signal.default_int_handler)

    @asyncio.coroutine
    def interact(self, banner=None, stop=True, handle_sigint=True):
        try:
            if self.streams is not None:
                self.reader, self.writer = yield from self.streams
        finally:
            self.streams = None
        try:
            if handle_sigint:
                self.add_sigint_handler()
            yield from self._interact(banner)
        except SystemExit:
            if stop:
                raise
        finally:
            if handle_sigint:
                self.remove_sigint_handler()
            if stop:
                self.loop.stop()

    @asyncio.coroutine
    def _interact(self, banner=None):
        # Get ps1 and ps2
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        # Print banner
        if banner is None:
            banner = self.get_default_banner()
        self.write("%s\n" % str(banner))
        # Run loop
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = yield from self.raw_input(prompt)
                except EOFError:
                    self.write("\n")
                    yield from self.flush()
                    break
                else:
                    more = yield from self.push(line)
            except asyncio.CancelledError:
                self.write("\nKeyboardInterrupt\n")
                yield from self.flush()
                self.resetbuffer()
                more = 0

    @asyncio.coroutine
    def push(self, line):
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        more = yield from self.runsource(source, self.filename)
        if not more:
            self.resetbuffer()
        return more

    @asyncio.coroutine
    def raw_input(self, prompt=''):
        self.write(prompt)
        yield from self.flush()
        data = (yield from self.reader.readline())
        try:
            data = data.decode()
        except UnicodeDecodeError:
            if b'\xff\xf4\xff\xfd\x06' in data:
                raise SystemExit
            data = '\n'
        if not data.endswith('\n'):
            raise EOFError
        return data.rstrip('\n')

    def write(self, data):
        return self.writer.write(data.encode())

    @asyncio.coroutine
    def flush(self):
        try:
            yield from self.writer.drain()
        except ConnectionResetError:
            pass

    # Re-implement showtraceback and showsyntaxerror
    # to ignore sys.excepthook (set by ubuntu apport for instance)

    def showtraceback(self):
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)
            self.write(''.join(lines))
        finally:
            last_tb = ei = None

    def showsyntaxerror(self, filename=None):
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        lines = traceback.format_exception_only(type, value)
        self.write(''.join(lines))



@asyncio.coroutine
def interact(banner=None, streams=None, locals=None, stop=True,
             handle_sigint=True, *, loop=None):
    console = AsynchronousConsole(streams, locals, loop=loop)
    yield from console.interact(banner, stop, handle_sigint)



"""Serve the python console using socket communication."""

import asyncio

from . import code


@asyncio.coroutine
def handle_connect(reader, writer, factory, banner=None):
    streams = reader, writer
    interface = factory(streams=streams)
    yield from interface.interact(banner=banner, stop=False,
                                  handle_sigint=False)
    writer.close()


@asyncio.coroutine
def start_interactive_server(factory=code.AsynchronousConsole,
                             host='localhost', port=8000, banner=None,
                             *, loop=None):
    callback = lambda reader, writer: handle_connect(
        reader, writer, factory, banner)
    server = yield from asyncio.start_server(callback, host, port, loop=loop)
    return server


@asyncio.coroutine
def start_console_server(host='localhost', port=8000,
                         locals=None, filename="<console>", banner=None,
                         *, loop=None):
    factory = lambda streams: code.AsynchronousConsole(
        streams, locals, filename)
    server = yield from asyncio.start_console_server(
        factory, host, port, banner, loop=loop)
    return server



def print_server(server, name='console'):
    interface = '{}:{}'.format(*server.sockets[0].getsockname())
    print('The {} is being served on {}'.format(name, interface))


def run(host='localhost', port=8000):
    loop = asyncio.get_event_loop()
    coro = start_interactive_server(host=host, port=port)
    loop.server = loop.run_until_complete(coro)
    print_server(loop.server)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass


def parse_server(server, parser=None):
    try:
        host, port = server.split(':')
    except ValueError:
        host, port = 'localhost', server
    try:
        port = int(port)
    except (ValueError, TypeError):
        msg = "{!r} is not a valid server [HOST:]PORT".format(server)
        if not parser:
            raise ValueError(msg)
        parser.error(msg)
    return host, port



#!/usr/bin/env python3

import sys
from setuptools import setup

TESTING = any(x in sys.argv for x in ['test', 'pytest'])

README = open("README.rst").read()

CLASSIFIERS = """\
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
""".splitlines()

setup(
    name="aioconsole",
    version="0.1.5",
    packages=["aioconsole"],
    entry_points={'console_scripts': ['apython = aioconsole:run_apython']},

    setup_requires=['pytest-runner' if TESTING else ''],
    tests_require=['pytest', 'pytest-asyncio<0.6', 'pytest-cov'],

    license="GPLv3",
    classifiers=CLASSIFIERS,
    description="Asynchronous console and interfaces for asyncio",
    long_description=README,

    author="Vincent Michel",
    author_email="vxgmichel@gmail.com",
    url='https://github.com/vxgmichel/aioconsole',
    download_url='http://pypi.python.org/pypi/aioconsole',
)



"""Provide asynchronous equivalent to input, exec and interact.

It also includes an interactive event loop, and a command line interface.
"""

from .execute import aexec
from .code import AsynchronousConsole, interact
from .stream import ainput, get_standard_streams
from .events import InteractiveEventLoop, InteractiveEventLoopPolicy
from .events import set_interactive_policy, run_console
from .command import AsynchronousCli
from .server import start_interactive_server
from .apython import run_apython

__all__ = ['aexec', 'ainput', 'AsynchronousConsole', 'interact',
           'InteractiveEventLoop', 'InteractiveEventLoopPolicy',
           'set_interactive_policy', 'run_console',
           'AsynchronousCli', 'start_interactive_server',
           'get_standard_streams', 'run_apython']
